from . layout import Layout
from . geometry.circle import (
    make_circle_coord_map, calc_ring_steps, calc_ring_pixel_count,
    make_circle_coord_map_positions)
from .. util import deprecated


class Circle(Layout):
    CLONE_ATTRS = Layout.CLONE_ATTRS + (
        'rings', 'pixels_per', 'maxAngleDiff', 'rotation', 'reverse_angle')

    def __init__(self, drivers, rings=[], pixels_per=None,
                 maxAngleDiff=0, rotation=0, reverse_angle=False,
                 threadedUpdate=False, brightness=255, **kwargs):
        super().__init__(drivers, threadedUpdate, brightness, **kwargs)
        self.rings = rings
        self.maxAngleDiff = maxAngleDiff
        full_coords = False
        for r in self.rings:
            full_coords |= (len(r) > 2)

        if not full_coords:
            self.rings, self.ringSteps = make_circle_coord_map(
                rings=self.rings, pixels_per=pixels_per)
        else:
            self.ringSteps = calc_ring_steps(self.rings)

        self.ringCount = len(self.rings)
        self.set_pixel_positions(make_circle_coord_map_positions(self.rings))

        num = calc_ring_pixel_count(self.rings)

        self.rotation = rotation
        self.reverse_angle = reverse_angle

        if self.numLEDs != num:
            raise ValueError(
                "Total ring LED count does not equal driver LED count!")

    @property
    def lastRing(self):
        deprecated.deprecated('Circle.lastRing')
        return self.ringCount - 1

    @property
    def shape(self):
        """Returns ``ringCount, ringSteps``."""
        return self.ringCount, self.ringSteps

    def __genOffsetFromAngle(self, angle, ring):
        if ring >= self.ringCount:
            return -1

        angle = (angle + self.rotation) % 360

        if self.reverse_angle:
            angle = 360 - angle

        offset = int(round(angle / self.ringSteps[ring]))

        # wraps it back around
        length = len(self.rings[ring]) - 1

        if offset > length:
            offset = 0

        if self.maxAngleDiff > 0:
            calcAngle = (offset * self.ringSteps[ring]) % 360
            diff = abs(angle - calcAngle)
            if diff > self.maxAngleDiff:
                return -1

        return offset

    def angleToPixel(self, angle, ring):
        offset = self.__genOffsetFromAngle(angle, ring)
        if offset < 0:
            return offset

        return self.rings[ring][offset]

    # Set single pixel to Color value
    def set(self, ring, angle, color):
        """Set pixel to RGB color tuple"""
        pixel = self.angleToPixel(angle, ring)
        self._set_base(pixel, color)

    def get(self, ring, angle):
        """Get RGB color tuple of color at index pixel"""
        pixel = self.angleToPixel(angle, ring)
        return self._get_base(pixel)

    def drawRadius(self, angle, color, startRing=0, endRing=-1):
        if startRing < 0:
            startRing = 0
        if endRing < 0 or endRing > self.ringCount - 1:
            endRing = self.ringCount - 1
        for ring in range(startRing, endRing + 1):
            self.set(ring, angle, color)

    def fillRing(self, ring, color, startAngle=0, endAngle=None):
        if endAngle is None:
            endAngle = 359

        if ring >= self.ringCount:
            raise ValueError("Invalid ring!")

        def pixelRange(ring, start=0, stop=-1):
            r = self.rings[ring]
            if stop == -1:
                stop = len(r) - 1

            return [r[i] for i in range(start, stop + 1)]

        start = self.__genOffsetFromAngle(startAngle, ring)
        end = self.__genOffsetFromAngle(endAngle, ring)
        if self.reverse_angle:
            start, end = end, start

        pixels = []
        if start >= end:
            pixels = pixelRange(ring, start)
            pixels.extend(pixelRange(ring, 0, end))
        elif start == end and startAngle > endAngle:
            pixels = pixelRange(ring)
        else:
            pixels = pixelRange(ring, start, end)

        for i in pixels:
            self._set_base(i, color)


if deprecated.allowed():  # pragma: no cover
    LEDCircle = Circle



from . circle import (
    calc_ring_pixel_count, calc_ring_steps, make_circle_coord_map,
    make_circle_coord_map_positions)

from . cube import make_cube_coord_map, make_cube_coord_map_positions

from . matrix import (
    make_matrix_coord_map, make_matrix_coord_map_multi,
    make_matrix_coord_map_positions)

from . strip import (
    make_strip_coord_map, make_strip_coord_map_multi,
    make_strip_coord_map_positions)

from ... util import deprecated
if deprecated.allowed():  # pragma: no cover
    from . rotation import Rotation, rotate_and_flip



def flatten(master):
    """
    :param dict master: a multilevel dictionary
    :return: a flattened dictionary
    :rtype: dict

    Flattens a multilevel dictionary into a single-level one so that::

        {'foo':
            {'bar':
               {
                   'a': 1,
                   'b': True,
                   'c': 'hello',
                },
            },
        }

    would become::

        {'foo.bar.a': 1,
         'foo.bar.b': True,
         'foo.bar.a': 1,
         }

    You can mix and match both input (hierarchical) and output (dotted) formats
    in the input without problems - and if you call flatten more than once, it
    has no effect.
    """
    result = {}

    def add(value, *keys):
        if keys in result:
            raise ValueError('Duplicate key %s' % keys)
        result[keys] = value

    def recurse(value, *keys):
        if isinstance(value, dict):
            for k, v in value.items():
                recurse(v, k, *keys)
        else:
            key = '.'.join(reversed(keys))
            if key in result:
                raise ValueError('Duplicate key %s' % str(keys))
            result[key] = value

    recurse(master)
    return result


def unflatten(master):
    """
    :param dict master: a multilevel dictionary
    :return: a unflattened dictionary
    :rtype: dict

    Unflattens a single-level dictionary a multilevel into one so that::

        {'foo.bar.a': 1,
         'foo.bar.b': True,
         'foo.bar.a': 1,
         }

    would become::

        {'foo':
            {'bar':
               {
                   'a': 1,
                   'b': True,
                   'c': 'hello',
                },
            },
        }
    """
    result = {}

    for k, v in master.items():
        *first, last = k.split('.')
        r = result
        for i in first:
            r = r.setdefault(i, {})
        r[last] = v

    return result


def canonical(master):
    """
    :param dict master: a multilevel dictionary
    :return: a canonicalized dictionary that has been completely flattened
             and then unflattened
    :rtype: dict
    """

    return unflatten(flatten(master))



import copy
from . wrapper import Wrapper
from .. util import color_list


class Feedback(Wrapper):
    def __init__(self, *args, animation,
                 master=1, inputs=None, outputs=None, **kwds):
        super().__init__(*args, animation=animation, **kwds)
        self.master = master

        inputs = inputs or [0]
        outputs = outputs or []

        in_sources = [copy.deepcopy(self.color_list) for i in inputs[1:]]
        in_sources.insert(self.animation.color_list)

        out_sources = [copy.deepcopy(self.color_list) for i in outputs]

        self.inputs = color_list.Mixer(self.color_list, in_sources, inputs)
        self.outputs = color_list.Mixer(self.color_list, out_sources, outputs)

        self.clear = self.inputs.clear
        self.math = self.inputs.math

    def step(self, amt=1):
        super().step(amt=amt)

        self.clear()
        self.inputs.mix(self.master)
        self.outputs.mix(self.master)

        def rotate(sources, begin):
            if len(sources) > 1 + begin:
                sources.insert(begin, sources.pop())

        ins, outs = self.inputs.sources, self.outputs.sources

        rotate(ins, 1)
        rotate(outs, 0)

        if len(ins) > 1:
            self.math.copy(ins[1], ins[0])

        if len(outs) > 0:
            self.math.copy(outs[0], self.color_list)



"""
Run specified project from file or URL
"""

import os, string, sys, time, traceback
from . import common_flags
from .. util import data_file, log, pid_context, signal_handler
from .. animation import Animation
from .. project import load
from .. project.aliases import ALIAS_MARKERS
from .. project.project import Project

RUN_ERROR = """When reading file {filename}:

{desc}
{exception}
"""

FAILURE_ERROR = """\
{count} project{s} failed
____________________
"""


def _load_py(filename):
    module = load.module(filename)
    project = module.get('PROJECT', {})

    if 'animation' not in project:
        no_extension = os.path.splitext(filename)[0]
        module_name = os.path.split(no_extension)[-1]
        names = module.keys()
        try:
            animation_name = load.guess_name(names, module_name, filename)
        except:
            if 'Animation' not in names:
                raise ValueError('Cannot deduce animation in file ' + filename)
            animation_name = 'Animation'
        animation = module[animation_name]
        if not callable(animation):
            raise ValueError('Animation "%s" in file "%s" is not callable'
                             % (animation_name, filename))

        project['animation'] = {'datatype': animation}

    return project


def _dump(args, desc):
    return data_file.dumps(desc, use_yaml=not args.json, safe=False)


def _get_projects(args):
    projects, failed = [], []

    gif = args.gif
    if gif != '':
        log.info('writing animated gifs')
        if gif is None:
            # --get appeared but without a value
            gif = '{}'
        if '{' in gif:
            gif = data_file.loads(gif, filename='--gif')
        if isinstance(gif, str):
            gif = {'filename': gif}
        if 'filename' in gif and len(args.name) > 1:
            log.warning('Writing multiple documents to the same GIF file name')

        gif['typename'] = '.gif_writer'

    # Handle spaces around + signs.
    names = []
    for name in args.name:
        if names and (name.startswith('+') or names[-1].endswith('+')):
            names[-1] += name
        else:
            names.append(name)
    args.name = names

    for project_files in args.name:
        saved_path = sys.path[:]
        descs = []
        try:
            for filename in project_files.split('+'):
                root_file = None
                if filename.endswith('.py'):
                    desc = _load_py(filename)
                elif '{' in filename:
                    desc = data_file.loads(filename)
                else:
                    try:
                        desc = load.data(filename, False)
                        desc = data_file.load(desc, filename)
                        root_file = os.path.abspath(filename)
                    except:
                        if not _might_be_classname(filename):
                            raise
                        desc = {'animation': filename}

                if gif:
                    if 'filename' in gif:
                        desc['driver'] = gif
                    else:
                        f, suffix = os.path.splitext(filename)
                        desc['driver'] = dict(gif, filename=f + '.gif')
                descs.append(desc)

            project = common_flags.make_project(
                args, *descs, root_file=root_file)
            projects.append(project)
            if args.dump:
                log.printer(_dump(args, project.desc))

        except FileNotFoundError as e:
            failed.append(('%s: %s' % (e.strerror, e.filename), ()))

        except Exception as exception:
            if filename.endswith('.py'):
                raise
            eargs = exception.args
            desc = _dump(args, descs)
            if args.verbose:
                exception = traceback.format_exc()
            msg = RUN_ERROR.format(**locals())
            failed.append((msg, eargs))

        finally:
            sys.path[:] = saved_path

    if not failed:
        return projects

    log.error(FAILURE_ERROR.format(
        count=len(failed), s='' if len(failed) == 1 else 's'))
    for msg, args in failed:
        if args:
            log.error(msg + '\n' + '\n'.join(str(a) for a in args) + '\n')
        else:
            log.error(msg + '\n')
    raise ValueError('Run aborted')


def _run_projects(projects, args):
    global _RUNNING
    _RUNNING = True

    needs_pause = False
    assert len(projects) == len(args.name)
    is_gif = args.gif != ''

    for project, name in zip(projects, args.name):
        if not _RUNNING:
            break

        if needs_pause:
            if args.pause:
                time.sleep(float(args.pause))
                if not _RUNNING:
                    break
        else:
            needs_pause = True

        log.debug('Running file %s', name)
        if is_gif:
            project.flat_out()

        try:
            project.run()

        except KeyboardInterrupt:
            if not args.ignore_exceptions:
                raise
            log.warning('\nKeyboardInterrupt terminated project.')
            needs_pause = False

        except Exception as e:
            if not args.ignore_exceptions:
                raise
            log.error('Exception %s', e)
            traceback.print_exc()


def run_once(args):
    Animation.FAIL_ON_EXCEPTION = args.fail_on_exception

    args.name = args.name or ['']
    projects = _get_projects(args)

    if args.dry_run:
        log.printer('(dry run - nothing executed)')
        return

    _run_projects(projects, args)


def stop():
    global _RUNNING
    _RUNNING = False
    Project.stop_all()


def run(args):
    with pid_context.pid_context(args.pid_filename):
        with signal_handler.context(
                SIGHUP=stop, SIGINT=stop, SIGTERM=stop) as signals:
            while True:
                run_once(args)
                if signals:
                    log.info('Received signal %s', ' '.join(signals))

                if not signals.pop('SIGHUP', False):
                    break


def set_parser(parser):
    parser.set_defaults(run=run)
    common_flags.add_project_flags(parser)

    parser.add_argument(
        'name', nargs='*',
        help='Path project files - can be a URL or file system location')


_RUNNING = False
_ALL_CHARS = set(string.ascii_letters + string.digits + '._' + ALIAS_MARKERS)
_START_CHARS = set(string.ascii_letters + '.' + ALIAS_MARKERS)
_END_CHARS = set(string.ascii_letters + string.digits)


def _might_be_classname(s):
    return (s and
            set(s) <= _ALL_CHARS and
            s[0] in _START_CHARS and
            s[-1] in _END_CHARS)



"""
Set some or all sections of the project defaults
"""

DESCRIPTION = """
Each argument to `bp default set` has to look like

section=value

Example:

    bp set numbers=float shape=80

"""

import sys
from ... project import merge, defaults


def run(args):
    defaults.set_defaults(args.sections)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('sections', nargs='*')



"""
Reset sections in the project defaults
"""

DESCRIPTION = """
Example:

``````
# Reset the layout and animation sections
$ bp default reset layout animation

# Reset all sections
$ bp default reset
``````
"""

from ... project import defaults


def run(args):
    defaults.reset_defaults(args.sections)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('sections', nargs='*')



import math


def generate_header(cmd, size):
    packet = bytearray()
    packet.append(cmd)
    packet.append(size & 0xFF)
    packet.append(size >> 8)
    return packet


def tuple_add(a, b):
    return tuple(x + y for x, y in zip(a, b))


def tuple_sub(a, b):
    return tuple(x - y for x, y in zip(a, b))


def tuple_mult(a, b):
    return tuple(x * y for x, y in zip(a, b))


def tuple_div(a, b):
    return tuple(x / y for x, y in zip(a, b))


def even_dist(start, stop, steps):
    steps -= 1
    start = float(start)
    stop = float(stop)
    div = (stop - start) / steps
    return [int(round(start + x * div)) for x in range(steps)] + [int(stop)]


def pointOnCircle(cx, cy, radius, angle):
    """
    Calculates the coordinates of a point on a circle given the center point,
    radius, and angle.
    """
    angle = math.radians(angle) - (math.pi / 2)
    x = cx + radius * math.cos(angle)
    if x < cx:
        x = math.ceil(x)
    else:
        x = math.floor(x)

    y = cy + radius * math.sin(angle)

    if y < cy:
        y = math.ceil(y)
    else:
        y = math.floor(y)

    return (int(x), int(y))


def genVector(width, height, x_mult=1, y_mult=1):
    """
    Generates a map of vector lengths from the center point to each coordinate.

    width - width of matrix to generate
    height - height of matrix to generate
    x_mult - value to scale x-axis by
    y_mult - value to scale y-axis by
    """
    center_x = (width - 1) / 2
    center_y = (height - 1) / 2

    def length(x, y):
        dx = math.pow(x - center_x, 2 * x_mult)
        dy = math.pow(y - center_y, 2 * y_mult)
        return int(math.sqrt(dx + dy))

    return [[length(x, y) for x in range(width)] for y in range(height)]



from . driver_base import DriverBase
from .. util.image import gif_writer as _gif_writer


class GifWriter(DriverBase):
    """
    An animation that writes animated GIFs for each frame in the contained
    animation."""

    def __init__(self, *args, filename='gif_writer', render=None,
                 divide=1, frames=0, time=10, scale=1.0, gif_options=None,
                 tmp_dir=None, **kwds):
        """
        :param str filename: Base filename to write the animated GIF file

        :param dict render: Parameters to the renderer function -
            see ``bibliopixel.util.image.render.renderer``

        :param int divide: If greater than 1, only rendered one in ``divide``
            frames

        :param int frames: Number of frames to write

        :param float time: Total time to write.  If non-zero, takes precedence
            over `frames`

        :param float speed: the length of the GIF is scaled by this factor, so
            if speed=2 then a 1 second animation will become a 2 second GIF.

        :param dict gif_options: Options to
            ``bibliopixel.util.image.gif.write_animation``

        :param str tmp_dir: If set, write individual GIF frame files to this
            directory, and do not delete them when done.  For testing purposes.
        """
        # TODO: this is duplicated here so that it's duplicated in the
        # documentation.  Is there a better way to do this?
        super().__init__(*args, **kwds)
        self.cur_step = 1
        self.gif_writer = _gif_writer.GifWriter(
            filename, render, divide, frames, time, scale, gif_options, tmp_dir)

    def set_project(self, project):
        super().set_project(project)
        self.gif_writer.set_project(project)

    def update_colors(self):
        super().update_colors()
        if self.gif_writer.step(self.cur_step):
            self.project.stop()
        else:
            self.cur_step += 1

    def cleanup(self):
        self.gif_writer.write()



import numbers


def crop(image, top_offset=0, left_offset=0, bottom_offset=0, right_offset=0):
    """Return an image cropped on top, bottom, left or right."""
    if bottom_offset or top_offset or left_offset or right_offset:
        width, height = image.size
        box = (left_offset, top_offset,
               width - right_offset, height - bottom_offset)
        image = image.crop(box=box)

    return image


def resize(image, x, y, stretch=False, top=None, left=None, mode='RGB',
           resample=None):
    """Return an image resized."""
    if x <= 0:
        raise ValueError('x must be greater than zero')
    if y <= 0:
        raise ValueError('y must be greater than zero')

    from PIL import Image

    resample = Image.ANTIALIAS if resample is None else resample
    if not isinstance(resample, numbers.Number):
        try:
            resample = getattr(Image, resample.upper())
        except:
            raise ValueError("(1) Didn't understand resample=%s" % resample)
        if not isinstance(resample, numbers.Number):
            raise ValueError("(2) Didn't understand resample=%s" % resample)

    size = x, y
    if stretch:
        return image.resize(size, resample=resample)
    result = Image.new(mode, size)

    ratios = [d1 / d2 for d1, d2 in zip(size, image.size)]
    if ratios[0] < ratios[1]:
        new_size = (size[0], int(image.size[1] * ratios[0]))
    else:
        new_size = (int(image.size[0] * ratios[1]), size[1])

    image = image.resize(new_size, resample=resample)
    if left is None:
        box_x = int((x - new_size[0]) / 2)
    elif left:
        box_x = 0
    else:
        box_x = x - new_size[0]

    if top is None:
        box_y = int((y - new_size[1]) / 2)
    elif top:
        box_y = 0
    else:
        box_y = y - new_size[1]

    result.paste(image, box=(box_x, box_y))
    return result



from . circle import Circle
from . matrix import Matrix
from . cube import Cube
from . pov import POV
from . strip import Strip
from . geometry.matrix import make_matrix_coord_map
from . geometry.circle import make_circle_coord_map
from . geometry.cube import make_cube_coord_map

from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    from . circle import LEDCircle
    from . cube import LEDCube
    from . matrix import LEDMatrix
    from . pov import LEDPOV
    from . strip import LEDStrip
    from . geometry.rotation import Rotation



"""These are operations that transform an x, y matrix index."""


def reflect_x(x, y, matrix):
    """Reflect the index vertically."""
    return matrix.columns - 1 - x, y


def reflect_y(x, y, matrix):
    """Reflect the index horizontally."""
    return x, matrix.rows - 1 - y


def serpentine_x(x, y, matrix):
    """Every other row is indexed in reverse."""
    if y % 2:
        return matrix.columns - 1 - x, y
    return x, y


def serpentine_y(x, y, matrix):
    """Every other column is indexed in reverse."""
    if x % 2:
        return x, matrix.rows - 1 - y
    return x, y


def transpose(x, y, _):
    """Transpose rows and columns."""
    return y, x



from . import strip


class Segment(strip.Strip):
    """Represents an offset, length segment within a strip."""

    def __init__(self, strip, length, offset=0):
        if offset < 0 or length < 0:
            raise ValueError('Segment indices are non-negative.')

        if offset + length > len(strip):
            raise ValueError('Segment too long.')

        self.strip = strip
        self.offset = offset
        self.length = length

    def __getitem__(self, index):
        return self.strip[self._fix_index(index)]

    def __setitem__(self, index, value):
        self.strip[self._fix_index(index)] = value

    def __len__(self):
        return self.length

    def next(self, length):
        """Return a new segment starting right after self in the same buffer."""
        return Segment(self.strip, length, self.offset + self.length)

    def _fix_index(self, index):
        if isinstance(index, slice):
            raise ValueError('Slicing segments not implemented.')
        if index < 0:
            index += self.length
        if index >= 0 and index < self.length:
            return self.offset + index
        raise IndexError('Index out of range')


def make_segments(strip, length):
    """Return a list of Segments that evenly split the strip."""
    if len(strip) % length:
        raise ValueError('The length of strip must be a multiple of length')

    s = []
    try:
        while True:
            s.append(s[-1].next(length) if s else Segment(strip, length))
    except ValueError:
        return s



from . import index_ops
from . rotation import rotate_and_flip
import copy


class Matrix(object):
    """Map a matrix onto a strip of lights."""

    def __init__(self, strip, columns=None, rows=None,
                 reflect_x=False, reflect_y=False,
                 serpentine_x=False, serpentine_y=False,
                 transpose=False):
        self.strip = strip

        if columns:
            if not rows:
                rows = len(strip) // columns
            elif rows * columns > len(strip):
                raise IndexError('Index out of range.')
        elif rows:
            columns = len(strip) // rows
        else:
            raise ValueError('Must supply either rows or columns')
        self.columns, self.rows = columns, rows

        ops = (reflect_x and index_ops.reflect_x,
               reflect_y and index_ops.reflect_y,
               serpentine_x and index_ops.serpentine_x,
               serpentine_y and index_ops.serpentine_y,
               transpose and index_ops.transpose)
        self.operations = list(filter(None, ops))

    def _index(self, x, y):
        for o in self.operations:
            x, y = o(x, y, self)

        if 0 <= x < self.columns and 0 <= y < self.rows:
            return x + y * self.columns

        raise IndexError('Index out of range.')

    def get(self, x, y):
        return self.strip[self._index(x, y)]

    def set(self, x, y, value):
        self.strip[self._index(x, y)] = value


def make_matrix_coord_map(
        dx, dy, serpentine=True, offset=0, rotation=0, y_flip=False):
    """Helper method to generate X,Y coordinate maps for strips"""
    result = []
    for y in range(dy):
        if not serpentine or y % 2 == 0:
            result.append([(dx * y) + x + offset for x in range(dx)])
        else:
            result.append([dx * (y + 1) - 1 - x + offset for x in range(dx)])

    result = rotate_and_flip(result, rotation, y_flip)

    return result


DEFAULT_CONFIG = {
    'serpentine': True,
    'rotation': 0,
    'y_flip': False
}


def make_matrix_coord_map_multi(config, rotation=0, y_flip=False):
    matrix_offset = 0
    matrix_result = []

    def populate_config(cfg):
        return dict(DEFAULT_CONFIG, **cfg)

    def add_row(offset, matrix, row):
        for y in row:
            matrix.append([x + offset for x in y])

        return offset + sum([len(y) for y in row])

    def combine_cols(cols):
        row_offset = 0
        row = []
        for col in cols:
            count = 0
            for i, y in enumerate(col):
                offset_col = [x + row_offset for x in y]
                if i >= len(row):
                    row.append(offset_col)
                else:
                    row[i] += offset_col
                count += len(col)
            row_offset += count
        return row

    if isinstance(config, dict):
        matrix_result = make_matrix_coord_map(**config)

    else:
        for row in config:
            if isinstance(row, dict):  # is configs
                row = populate_config(row)
                matrix_offset = add_row(
                    matrix_offset, matrix_result, make_matrix_coord_map(**row))

            elif isinstance(row, list):  # is row
                cols = []
                for config in row:
                    config = populate_config(config)
                    cols.append(make_matrix_coord_map(**config))

                matrix_offset = add_row(
                    matrix_offset, matrix_result, combine_cols(cols))

    return rotate_and_flip(matrix_result, rotation, y_flip)


def make_matrix_coord_map_positions(coord_map):
    num = len(coord_map) * max(len(x) for x in coord_map)
    result = [None] * num

    for y in range(len(coord_map)):
        for x in range(len(coord_map[y])):
            result[coord_map[y][x]] = [x, y, 0]

    return result



import os, tempfile
from .. project import data_maker, defaults, project
from .. util import data_file, deprecated, log, pid_context

"""Common command line arguments for run and demo."""


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(dirname(dirname(abspath(
        __file__)))), 'bibliopixel', 'VERSION')
    return open(filename).read().strip()


VERSION = _get_version()
COMPONENTS = 'driver', 'layout', 'animation'
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
ENABLE_PRESETS = False
NUMBER_TYPES = ('python',) + data_maker.NUMPY_TYPES


def add_common_flags(parser):
    parser.add_argument(
        '--loglevel', choices=log.SORTED_NAMES, default='info',
        help=LOGLEVEL_HELP)
    parser.add_argument(
        '--pid_filename', default=pid_context.DEFAULT_PID_FILENAME,
        help='Filename to store the `bp` process ID when running')
    parser.add_argument(
        '--verbose', '-v', action='store_true', help=VERBOSE_HELP)
    parser.add_argument(
        '--version', action='store_true', help=VERSION_HELP)
    parser.add_argument(
        deprecated.V4_FLAG, action='store_true', help=V4_HELP)


def execute_args(args):
    if args.verbose and args.loglevel != 'frame':
        log.set_log_level('debug')
    else:
        log.set_log_level(args.loglevel)


def add_project_flags(parser):
    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')

    parser.add_argument(
        '-b', '--brightness', default=None,
        help='Override project brightness value')

    parser.add_argument(
        '-d', '--defaults', default=None, nargs='*',
        action='append', help='Use this default setting')

    if deprecated.allowed():  # pragma: no cover
        parser.add_argument(
            '--dimensions', '--dim', default=None,
            help='DEPRECATED: x, (x, y) or (x, y, z) dimensions for project')

    parser.add_argument(
        '--shape', default=None,
        help='x, (x, y) or (x, y, z) dimensions for project')

    parser.add_argument(
        '--dump', action='store_true',
        help='Dump the full project as DATA_FILE after loading but ' +
             'before running')

    parser.add_argument(
        '--dry_run', action='store_true',
        help='Load projects but do not run them')

    parser.add_argument(
        '-f', '--fail_on_exception', action='store_true',
        help='If true, bp fail if any subanimation fails to construct')

    parser.add_argument(
        '-g', '--gif', default='', nargs='?', help=GIF_HELP)

    parser.add_argument(
        '-i', '--ignore_exceptions', action='store_true',
        help='If true, continue running the next project if one project fails')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Use Json when dumping description data')

    parser.add_argument(
        '-l', '--layout', default=None,
        help='Default layout class if no layout is specified')

    parser.add_argument(
        '--numbers', '-n', default='python', choices=NUMBER_TYPES,
        help=NUMBERS_HELP)

    parser.add_argument('-p', '--path', default=None, help=PATH_HELP)

    parser.add_argument(
        '--pause', default=0, help='Time to pause between running animations')

    parser.add_argument(
        '--project_lengths', '--pl', default=None, help=PROJECT_LENGTHS_HELP)

    parser.add_argument(
        '-s', action='store_true', help='Run SimPixel at the default URL')

    parser.add_argument(
        '--simpixel', help='Run SimPixel at a specific URL')

    parser.add_argument(
        '--animation_lengths', '--at', default=None,
        help='Set run length for each animation')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')


def _make_project_flags(args):
    def get_value(name):
        value = getattr(args, name, None)
        if not value:
            return {}

        if '{' in value:
            return data_file.loads(value)

        return {'typename': value}

    project_flags = {name: get_value(name) for name in COMPONENTS}
    if args.ledtype:
        project_flags['driver']['ledtype'] = args.ledtype

    if args.brightness:
        project_flags['layout']['brightness'] = int(args.brightness)

    if args.defaults:
        # Arguments come in like: [['foo'], ['bar'], ['baz'], ...]
        # I have no idea why.
        defs = [(a[0] if isinstance(a, list) else a) for a in args.defaults]
        defaults.set_project_defaults(defs)

    if args.v4:
        log.printer('*** Using v4 forward compatibility mode.')
        project_flags['numbers'] = 'float'
    elif args.numbers != 'python':
        project_flags['numbers'] = args.numbers

    if args.project_lengths is not None:
        run = project_flags.setdefault('run', {})
        run['seconds'] = float(args.project_lengths)

    if args.animation_lengths is not None:
        animation = project_flags.setdefault('animation', {})
        length = [float(i) for i in args.animation_lengths.split(',')]
        animation['length'] = length

    if args.dimensions is not None:
        deprecated.deprecated('Use --shape: --dimensions')

    shape = args.shape or args.dimensions
    if shape is not None:
        shape = shape.split(',')
        try:
            project_flags['shape'] = [int(i) for i in shape]
        except:
            raise ValueError('--shape must be one to three numbers '
                             'separated by commas.')

    return project_flags


def make_project(args, *descs, root_file=None):
    project_flags = _make_project_flags(args)
    descs += (project_flags,)
    return project.project(*descs, root_file=root_file)


# Help messages.

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use loady-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

LOGLEVEL_HELP = """\
Set what level of events to log. Higher log levels print less."""

NUMBERS_HELP = """
The type of numbers that are used in color list calculations.

  `python` means to use the classic Python lists of (r, g, b) tuples.

  Anything else is a numpy type, which means that bp uses numpy arrays, which
  use much faster arithmetic.

  numpy types include:
    """ + ' '.join(data_maker.NUMPY_TYPES)

PRESET_HELP = """Filenames for preset library"""

PROJECT_LENGTHS_HELP = """\
How long to run the animation (overrides runner.seconds)"""

VERBOSE_HELP = """\
If this is set, then errors are reported with a full stack trace, and
loglevel is by default set to debug
"""

VERSION_HELP = """\
Print the current version number of BiblioPixel (%s)
""" % VERSION

V4_HELP = """\
Run BiblioPixel in v4 compatibility mode, to see if it will work with
future releases v4.x
"""

GIF_HELP = """
Write an animated GIF file.

If --gif has no argument, the name of the GIF file is the same as the name of
the project.

If it has a single string argument, then it's the name of the GIF file.

Otherwise, the argument is read as JSON and used to construct a GifWriter class.
"""

# TODO: this should go somewhere else
"""
Set defaults for projects using YAML or JSON data files.

A Project is a data file or equivalently a Python dictionary that describes
a BibliopPixel installation.  The top-level keys in a Project are called the
sections and a Project might have the following sections:

    {sections}

Often some sections of your project correspond to hardware and thus rarely
change, so the `bp default` command allows you to set defaults so you
don't have to mention these from your project at all.
"""



import functools
from ...drivers.spi_interfaces import SPI_INTERFACES

USAGE = """
A spi_interface is represented by a string.

Possible values are """ + ', '.join(sorted(SPI_INTERFACES.__members__))


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(SPI_INTERFACES)
def _(c):
    return c


@make.register(str)
def _(c):
    return SPI_INTERFACES[c]



from . import wrapper
from .. util import log
import random as rand


class Sequence(wrapper.Indexed):
    def __init__(self, layout, random=False, length=None, **kwds):
        """
        Arguments

        random -- If True, a random animation is selected each step
        length -- if length is a number, run all the animations in a loop
            for `length` seconds each.  If `length` is a list of numbers,
            use the numbers successively as times.
        """
        super().__init__(layout, **kwds)
        self.random = random
        if isinstance(length, (list, tuple)):
            self.length = length
        else:
            self.length = length and [length]

    def restart(self):
        self.random and rand.shuffle(self.animations)
        self.index = 0

    def pre_run(self):
        self.offset = 0
        super().pre_run()
        self.restart()

    def step(self, amt=1):
        if (not self.animations) or super().step(amt):
            return

        self.index += 1
        if self.index >= len(self.animations):
            if self.runner.until_complete:
                self.completed = True
            else:
                self.offset += len(self.animations)
                self.restart()

    def _on_index(self, old_index):
        if self.current_animation:
            super()._on_index(old_index)
            if self.length:
                step = self.offset + self.index
                length = self.length[step % len(self.length)]
                self.current_animation.runner.seconds = length



from ... util import colors, log
from ... layout import Matrix, Strip


def renderer(layout, color, pixel_width, pixel_height,
             ellipse, vertical, frame, padding):
    if not isinstance(layout, (Matrix, Strip)):
        raise ValueError('Cannot render a layout of type %s' % type(layout))

    from PIL import Image, ImageDraw

    shape = layout.shape
    if len(shape) == 1:
        if vertical:
            shape, getter = (1, shape[0]), lambda x, y: layout.get(y)
        else:
            shape, getter = (shape[0], 1), lambda x, y: layout.get(x)
    else:
        getter = layout.get

    # LED width and height
    width, height = shape

    # width and height for individual pixels
    pw, ph = pixel_width, pixel_width
    ph = pw if ph is None else ph

    # Cell width and height
    cw, ch = pw + 2 * padding, ph + 2 * padding
    image_size = 2 * frame + width * cw, 2 * frame + height * ch

    def render():
        image = Image.new('RGB', image_size, color)
        draw = ImageDraw.Draw(image)
        draw_pixel = draw.ellipse if ellipse else draw.rectangle
        offset = frame + padding

        for x in range(width):
            px = offset + x * cw
            for y in range(height):
                py = offset + y * ch
                pcolor = tuple(int(i) for i in getter(x, y))
                draw_pixel((px, py, px + pw, py + ph), pcolor, pcolor)

        return image

    return render



import math, threading, time
from . matrix import Matrix


# Takes a matrix and displays it as individual columns over time
class POV(Matrix):

    def __init__(self, drivers, povHeight, width, rotation=0,
                 vert_flip=False, threadedUpdate=False,
                 brightness=255, **kwargs):
        self.numLEDs = povHeight * width

        super().__init__(drivers, width, povHeight, None,
                         rotation, vert_flip, threadedUpdate, brightness,
                         **kwargs)

    # This is the magic. Overriding the normal push_to_driver() method
    # It will automatically break up the frame into columns spread over
    # frameTime (ms)
    def push_to_driver(self, frameTime=None):
        if frameTime:
            self.animation_sleep_time = frameTime

        sleep = None
        if self.animation_sleep_time:
            sleep = (self.animation_sleep_time -
                     self.frame_render_time) / self.width

        width = self.width
        for h in range(width):
            start = time.time()

            def color(i):
                return self._colors[(h + width * i) * 3]
            # TODO: Figure out what line below needs to do
            # buf = [color(i) for i in range(self.height)]
            self.drivers[0].update_colors()
            sendTime = time.time() - start
            if sleep:
                time.sleep(max(0, sleep - sendTime))


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    LEDPOV = POV



import collections
from . import construct, load

# Elements in the DEFAULT_PROJECT are listed in order of dependency: sections
# can only be dependent on sections that are above them in this list.
# For example, animation is at the bottom which are  dependent on run and
# layout, which depend on drivers, and everything depends on path and typename.
DEFAULT_PROJECT = collections.OrderedDict((
    ('aliases', {'bpa': 'BiblioPixelAnimations'}),
    ('path', ''),
    ('typename', 'bibliopixel.project.project.Project'),
    ('numbers', 'python'),
    ('maker', 'bibliopixel.project.data_maker.Maker'),
    ('driver', {}),
    ('drivers', []),
    ('shape', ()),
    ('layout', {}),
    ('run', {}),
    ('animation', {'typename': 'animation'}),
    ('controls', []),
))
PROJECT_SECTIONS = tuple(DEFAULT_PROJECT.keys())

NOT_MERGEABLE = (
    'controls', 'datatype', 'dimensions', 'shape', 'drivers', 'numbers', 'path',
    'typename')

SECTION_ISNT_DICT_ERROR = 'Project section "%s" is %s, should be dictionary'
UNKNOWN_SECTION_ERROR = 'There is no Project section named "%s"'


def merge(*projects):
    """
    Merge zero or more dictionaries representing projects with the default
    project dictionary and return the result
    """
    result = {}
    for project in projects:
        for name, section in (project or {}).items():
            if name in NOT_MERGEABLE:
                result[name] = section
                continue

            if name not in DEFAULT_PROJECT:
                raise ValueError(UNKNOWN_SECTION_ERROR % name)

            if section and not isinstance(section, (dict, str)):
                cname = section.__class__.__name__
                raise ValueError(SECTION_ISNT_DICT_ERROR % (name, cname))

            if name == 'animation':
                # Useful hack to allow you to load projects as animations.
                adesc = load.load_if_filename(section)
                if adesc:
                    section = adesc.get('animation', {})
                    section['run'] = adesc.get('run', {})

            result.setdefault(name, {}).update(construct.to_type(section))
            if 'datatype' in result[name]:
                result[name].pop('typename', None)

    return result



from ... util.colors import gamma
from .. channel_order import ChannelOrder
from . base import SPIBase


class APA102(SPIBase):
    """Driver for APA102/SK9822 based LED strips on devices like
    the Raspberry Pi and BeagleBone

    Provides the same parameters as
    :py:class:`bibliopixel.drivers.SPI.SPIBase`
    """

    def __init__(self, num, gamma=gamma.APA102, **kwargs):
        super().__init__(num, gamma=gamma, **kwargs)

        # APA102/SK9822 requires latch bytes at the end)
        # Many thanks to this article for combined APA102/SK9822 protocol
        # https://cpldcpu.com/2016/12/13/sk9822-a-clone-of-the-apa102/
        self._start_frame = 4  # start frame is [0, 0, 0, 0]
        self._pixel_bytes = self.numLEDs * 4  # 4 byte frames [bright, r, g, b]
        self._pixel_stop = self._start_frame + self._pixel_bytes
        self._reset_frame = 4  # for SK9822 [0, 0, 0, 0]
        self._end_frame = (num // 2) + 1
        self._packet = self.maker.bytes(self._start_frame + self._pixel_bytes +
                                        self._reset_frame + self._end_frame)

        self.set_device_brightness(0xFF)  # required to setup _packet

    def set_device_brightness(self, val):
        """
        APA102 & SK9822 support on-chip brightness control, allowing greater
        color depth.

        APA102 superimposes a 440Hz PWM on the 19kHz base PWM to control
        brightness. SK9822 uses a base 4.7kHz PWM but controls brightness with a
        variable current source.

        Because of this SK9822 will have much less flicker at lower levels.
        Either way, this option is better and faster than scaling in
        BiblioPixel.
        """
        # bitshift to scale from 8 bit to 5
        self._chipset_brightness = (val >> 3)
        self._brightness_list = [0xE0 + self._chipset_brightness] * self.numLEDs
        self._packet[self._start_frame:self._pixel_stop:4] = (
            self._brightness_list)

    def _compute_packet(self):
        self._render()
        self._packet[self._start_frame + 1:self._pixel_stop:4] = self._buf[0::3]
        self._packet[self._start_frame + 2:self._pixel_stop:4] = self._buf[1::3]
        self._packet[self._start_frame + 3:self._pixel_stop:4] = self._buf[2::3]



from . address import number, Address
from . receiver import Receiver


class Editor(Receiver):
    """
    A `Editor` is a `Receiver` which gets and sets `Address`es, perhaps using an
    an optional `EditQueue` for the setting.

    When the `set_root` method is called, `Editor` searches
    down through the address and stores the most recent `edit_queue`
    method it finds.
    """

    def __init__(self, address=None, root=None):
        self.address = Address(address)
        if root is not None:
            self.set_root(root)

    def set_root(self, root):
        self.root = root
        self.segments = self.address.segments[:]
        self.last_segment = self.address.segments.pop()
        edit_queue = None

        for segment in self.address.segments:
            edit_queue = getattr(root, 'edit_queue', edit_queue)
            root = segment.get(root)

        self.edit_queue = edit_queue

    def receive(self, msg):
        """
        Receives a message, and either sets it immediately, or puts it on the
        edit queue if there is one.

        """
        if self.edit_queue:
            self.edit_queue.put_edit(self._set, msg)
        else:
            self._set(msg)

    def get(self):
        return self.last_segment.get(self._get())

    def set(self, value):
        self.receive((number(value),))

    def __bool__(self):
        return bool(self.address)

    def __str__(self):
        return str(self.address)

    def _get(self):
        root = self.root
        for segment in self.address.segments:
            root = segment.get(root)
        return root

    def _set(self, values):
        args = self.address.assignment + values
        self.last_segment.set(self._get(), *args)



import functools
from . import color
from ... util import color_list

USAGE = """
Colors is a list of colors.  Each color can be:

* A list of three numbers: [0, 0, 0] or [255, 0, 255].
* A single number which represents a brightness/gray level: 0, 255, 127
* A string:  "red", "yellow", "gold" naming a color from ...colors.COLORS.

All numbers must be in the range [0, 256) - 0 <= x < 256"""


@functools.singledispatch
def make(c):
    raise ValueError("Don't understand type %s" % type(c), USAGE)


@make.register(tuple)
@make.register(list)
def _(c):
    return [color.make(i) for i in color_list.to_triplets(c)]


@make.register(str)
def _(s):
    return make([s])



"""
Send a kill signal to a BiblioPixel process running on this
machine to abruptly kill it

DEPRECATED: use

.. code-block:: bash

    $ kill -kill `bpa-pid`

"""

DESCRIPTION = """
Example:

.. code-block:: bash

    $ bp kill

"""

from . restart import run, CHOICES


def set_parser(parser):
    parser.add_argument(
        'signal', nargs='?', default='SIGKILL', choices=CHOICES,
        help='Signal to send.')

    parser.set_defaults(run=run)



import serial
from ... drivers.return_codes import RETURN_CODES
from ... util import log, util


def send_packet(cmd, size, dev, baudrate, timeout, more_data=None):
    packet = util.generate_header(cmd, size)
    if more_data:
        packet += more_data

    com = serial.Serial(dev, baudrate=baudrate, timeout=timeout)
    com.write(packet)
    code = read_byte(com)
    ok = (code == RETURN_CODES.SUCCESS)
    return com, code, ok


def read_byte(com):
    try:
        resp = com.read(1)
        if resp:
            return ord(resp)
    except Exception as e:
        log.error('Serial exception %s in read', e)



from .. channel_order import ChannelOrder
from .. driver_base import DriverBase
from ... util.colors import gamma as _gamma
from . import interfaces


class SPIBase(DriverBase):
    """Base driver for controling SPI devices on systems like the Raspberry Pi
      and BeagleBone

    Provides the same parameters of
    :py:class:`bibliopixel.drivers.driver_base.DriverBase` as
    well as those below:

    :param str dev: SPI device path
    :param str spi_interface: Interface API with which to connect to the SPI
        device. One of
        :py:class:`bibliopixel.drivers.spi_interfaces.SPI_INTERFACES`
    :param int spi_speed: Output data rate, in MHz
    :param interface: DEPRECATED - Use spi_interface
    """

    def __init__(self, num, dev='/dev/spidev0.0',
                 spi_interface=None, spi_speed=1,
                 interface=None, **kwargs):
        super().__init__(num, **kwargs)
        from ...project.types.spi_interface import make
        # See https://github.com/ManiacalLabs/BiblioPixel/issues/419
        if interface:
            from ... util import deprecated
            deprecated.deprecated('SPIBase.interface')
        else:
            interface = 'FILE'

        maker = interfaces._SPI_INTERFACES[make(spi_interface or interface)]
        self._interface = maker(dev=dev, spi_speed=spi_speed)

    def _send_packet(self):
        self._interface.send_packet(self._packet)

    def _compute_packet(self):
        self._render()
        self._packet = self._interface.compute_packet(self._buf)



from ... project import load


def run_trigger(typename, q, events, kwargs):
    trigger_class = load.code(typename)
    trigger = trigger_class(q, events, **kwargs)
    try:
        trigger.start()
    except KeyboardInterrupt:
        pass



import functools, numbers
from ... util import colors

USAGE = """
A Color can be initialized with:

* A list of three numbers: [0, 0, 0] or [255, 0, 255].
* A single number which represents a brightness/gray level: 0, 255, 127
* A string:  "red", "yellow", "gold" naming a color from ...colors.COLORS.

All numbers must be in the range [0, 256) - 0 <= x < 256"""


@functools.singledispatch
def make(c):
    raise ValueError('Don\'t understand color name "%s"' % c, USAGE)


@make.register(tuple)
def _(c):
    if len(c) != 3:
        raise ValueError('Length %d is not 3' % len(c), USAGE)

    if not all(0 <= x <= 255 for x in c):
        raise ValueError('All components must be between 0 and 255', USAGE)

    return c


@make.register(list)
def _(c):
    return make(tuple(c))


@make.register(numbers.Number)
def _(c):
    if not (0 <= c <= 255):
        raise ValueError('All components must be between 0 and 255', USAGE)
    return (c, c, c)


@make.register(str)
def _(c):
    try:
        return colors.COLORS[c]
    except:
        raise ValueError('Don\'t understand color name "%s"' % c, USAGE)



import argparse, sys
from .. util import log

ARGS = None


def parse_args(commands, common_flags):
    helps = '--help', 'help'
    argv = ['-h' if a in helps else a for a in sys.argv[1:]]

    if not argv:
        argv = ['-h']

    # argparse doesn't give command-specific help for `bp --help <command>`
    # so we use `bp <command> --help` (#429)
    if len(argv) == 2 and argv[0] == '-h':
        argv.reverse()

    try:
        argv.remove('--version')
    except:
        pass
    else:
        log.printer('BiblioPixel version %s' % common_flags.VERSION)
        if not argv:
            return

    # Move all the flags to the end.
    args = [], []
    for a in argv:
        args[a.startswith('-')].append(a)

    argv = args[0] + args[1]

    if not argv[0].isidentifier() and '-h' not in argv:
        # The first argument can't be a command so try to run it.
        argv.insert(0, 'run')

    if argv and argv[0].startswith('-') and any(
            not a.startswith('-') for a in argv):
        log.printer(
            'bibliopixel: error: command line flags must appear at the end.')

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    for name, module in sorted(commands.MODULES.items()):
        doc = module.__doc__
        subparser = subparsers.add_parser(name, help=doc)
        common_flags.add_common_flags(subparser)
        module.set_parser(subparser)
        description = getattr(module, 'DESCRIPTION', '')
        subparser.description = doc + description

    if argv == ['-h']:
        log.printer(commands.HELP)

    global ARGS
    ARGS = parser.parse_args(argv)



"""
Print information about BiblioPixel
"""

DESCRIPTION = """
Prints the versions of BiblioPixel's dependencies, and the platform
that the program is running on.
"""

import datetime, os, platform, sys
import bibliopixel, BiblioPixelAnimations, loady
from bibliopixel.util import log

NONE = '(none)'


def run(args):
    now = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')
    log.printer('Timestamp:     ', now)
    log.printer('Python version:', platform.python_version())
    log.printer('Platform:      ', platform.platform())
    log.printer('`bp` path:     ', sys.argv[0])
    log.printer('Library path:  ', os.path.dirname(bibliopixel.__file__))

    log.printer()
    log.printer('Dependencies:')
    for module in bibliopixel, BiblioPixelAnimations, loady:
        path = os.path.dirname(module.__file__)
        parent = os.path.dirname(path)

        try:
            fp = open(os.path.join(path, 'VERSION'))
        except:
            try:
                fp = open(os.path.join(parent, 'VERSION'))
            except:
                fp = None

        version = fp.read().strip() if fp else NONE

        try:
            import git
            repo = git.Repo(os.path.dirname(path))
        except:
            commit_id = tag = NONE
        else:
            commit_id = repo.commit('HEAD').hexsha[:7]
            tag = repo.tags[-1].name if repo.tags else '(none)'

        log.printer('    %s: version %s, git commit: %s, git tag %s' %
                    (module.__name__, version, commit_id, tag))


def set_parser(parser):
    parser.set_defaults(run=run)



import math, threading, time

from .. util import colors, log
from . import font
from . layout import Layout
from . geometry.cube import make_cube_coord_map, make_cube_coord_map_positions


class Cube(Layout):
    CLONE_ATTRS = Layout.CLONE_ATTRS + ('x', 'y', 'z')

    def __init__(self, drivers, x, y, z, coord_map=None,
                 threadedUpdate=False, brightness=255, **kwargs):
        super().__init__(drivers, threadedUpdate, brightness, **kwargs)
        self.x = x
        self.y = y
        self.z = z

        if self.x * self.y * self.z != self.numLEDs:
            raise TypeError("(x * y * z) MUST equal the total pixel count!")

        if coord_map:
            self.coord_map = coord_map
        else:
            if len(self.drivers) == 1:
                self.coord_map = make_cube_coord_map(x, y, z)
            else:
                raise TypeError(
                    "Must provide coord_map if using multiple drivers!")

        self.set_pixel_positions(make_cube_coord_map_positions(self.coord_map))

    def get_pixel_positions(self):
        return make_cube_coord_map_positions(self.coord_map)

    @property
    def shape(self):
        """Returns ``x, y, z``"""
        return self.x, self.y, self.z

    def set(self, x, y, z, color):
        try:
            pixel = self.coord_map[z][y][x]
            self._set_base(pixel, color)
        except IndexError:
            pass

    def get(self, x, y, z):
        try:
            pixel = self.coord_map[z][y][x]
            return self._get_base(pixel)
        except IndexError:
            return 0, 0, 0

    def setHSV(self, x, y, z, hsv):
        color = colors.hsv2rgb(hsv)
        self._set(x, y, z, color)

    def setRGB(self, x, y, z, r, g, b):
        color = (r, g, b)
        self._set(x, y, z, color)


from .. util import deprecated
if deprecated.allowed():  # pragma: no cover
    LEDCube = Cube



from .. import log
import contextlib, functools, threading, time, traceback, queue


class Runnable:
    """
    Base class for all objects that contain threads - including threads
    created and started by bp code, and threads started by external libraries.

    There are three possible thread categories for a Runnable:

    1. Runs on the master thread - M
    2. A new thread created by us - N
    3. A new external thread created by some third-party code - X

    Case X is tricky because we would like our code to be called at the start of
    the new thread, and again at the end of that thread, but we can't.

    Lifecycle - what a Runnable does, and what threads it could be called on

    * construction: M
    * start: M
    * on_start: M
      * called on the master thread after any new thread has started up
    * run: MN
    * callbacks: MNX
    * join: M
    * cleanup: M(N?)

    TODO: right now we run all our cleanups on the new thread, if there is a new
    thread, otherwise on the master thread.  Should we move to doing all the
    cleanups on the master thread?

    The way to use a Runnable is like a context manager:

    with some_runnable() as runnable:
         add_some_callbacks(runnable)
         more_stuff_that_runs_on_start()

         # Depending on the thread category, the `Runnable` isn't guaranteed to
         # actually "go off" until the end of this block.

    We're going to call the code inside the context manager `on_start`

    """

    MASTER, NEW, EXTERNAL = 'M', 'N', 'X'
    category = NEW

    timeout = 0.1

    def __init__(self):
        self.run_event = threading.Event()
        self.stop_event = threading.Event()

    @property
    def running(self):
        """
        Is this Runnable expected to make any progress from here?

        The Runnable might still execute a little code after it has stopped
        running.
        """
        return self.run_event.is_set() and not self.stop_event.is_set()

    def is_alive(self):
        """
        Is this Runnable still executing code?

        In some cases, such as threads, self.is_alive() might be true for some
        time after self.running has turned False.
        """
        return self.running

    def start(self):
        self.run_event.set()

    def stop(self):
        self.stop_event.set()

    def wait(self):
        self.stop_event.wait()

    def cleanup(self):
        """
        Cleans up resources after the Runnable.

        self.cleanup() may not throw an exception.
        """

    def run(self):
        try:
            while self.running:
                self.run_once()
        except:
            log.error('Exception at %s: \n%s',
                      str(self), traceback.format_exc())
        finally:
            self.stop()
            self.cleanup()

    def run_once(self):
        """The target code that is repeatedly executed in the run method"""
        pass

    @contextlib.contextmanager
    def run_until_stop(self):
        """
        A context manager that starts this Runnable, yields,
        and then waits for it to finish."""
        self.start()
        try:
            yield self
        finally:
            self.stop()
        self.wait()


class LoopThread(Runnable):
    def __init__(self, daemon=True, **kwds):
        super().__init__()
        self.thread = threading.Thread(daemon=daemon, target=self.run, **kwds)

    def start(self):
        self.thread.start()

    def run(self):
        super().start()
        super().run()

    def is_alive(self):
        return self.thread.is_alive()


class QueueHandler(LoopThread):
    def __init__(self, timeout=0.1, send=None, **kwds):
        super().__init__(**kwds)
        self.timeout = timeout
        self.queue = queue.Queue()
        self.send = send or self.send

    def run_once(self):
        try:
            msg = self.queue.get(timeout=self.timeout)
        except queue.Empty:
            pass
        else:
            self.send(msg)

    def send(self, msg):
        pass



from . layout import Circle, Matrix, POV, Strip, Cube, font
from . layout.geometry.matrix import make_matrix_coord_map
from . layout.geometry.circle import make_circle_coord_map
from . layout.geometry.cube import make_cube_coord_map
from . import animation, layout, util
from . util import colors, image, log

from . util.colors import gamma
from . drivers import return_codes
from . project import data_maker

from . util import deprecated
if deprecated.allowed():  # pragma: no cover
    from . layout import LEDCircle, LEDMatrix, LEDPOV, LEDStrip, LEDCube
    from . layout import matrix_drawing as matrix
    from . layout.geometry.rotation import Rotation


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(abspath(__file__)), 'VERSION')
    return open(filename).read().strip()


__version__ = _get_version()
VERSION = __version__



import time
from enum import IntEnum
from .. util import log
from .. project import attributes, load

DEFAULT_FPS = 24


class STATE(IntEnum):
    ready = 0
    running = 1
    complete = 2
    canceled = 3
    max_steps = 4
    timeout = 5


class Runner(object):

    def __init__(self, *, amt=1, fps=0, sleep_time=0, max_steps=0,
                 until_complete=False, max_cycles=0, seconds=None,
                 threaded=False, main=None, flat_out=False,
                 repeats=None, **kwds):
        attributes.check(kwds, 'run')

        if max_steps < 0:
            log.error('max_steps %s < 0', max_steps)
            max_steps = 0
        if sleep_time < 0:
            log.error('sleep_time %s < 0', sleep_time)
            sleep_time = 0
        if max_cycles < 0:
            log.error('max_cycles %s < 0', max_cycles)
            max_cycles = 0
        if fps < 0:
            log.error('fps %s < 0', fps)
            fps = 0
        if repeats and repeats < 0:
            log.error('repeats %s < 0', repeats)
            repeats = None

        if sleep_time and fps:
            log.error('sleep_time=%s and fps=%s cannot both be set',
                      sleep_time, fps)
            sleep_time = 0
        if seconds and max_steps:
            log.error('seconds=%s and max_steps=%s cannot both be set',
                      seconds, max_steps)
            max_steps = 0

        self.amt = amt

        if fps:
            self.sleep_time = 1 / fps
        elif sleep_time:
            self.sleep_time = sleep_time
        else:
            self.sleep_time = 1 / DEFAULT_FPS

        self.until_complete = until_complete
        self.seconds = seconds
        self.run_start_time = 0
        self.max_steps = max_steps
        self.max_cycles = max_cycles
        self.seconds = seconds
        self.threaded = threaded
        self.flat_out = flat_out
        self.main = load.code(main)
        if repeats is not None:
            self.until_complete = True
            self.max_cycles = repeats
        self.repeats = repeats

    def set_project(self, project):
        self.project = project
        if self.flat_out:
            project.flat_out()

    @property
    def fps(self):
        return 1 / self.sleep_time

    @fps.setter
    def fps(self, fps):
        self.sleep_time = 1 / fps

    def compute_state(self, cur_step, state):
        if self.seconds:
            elapsed = self.project.time() - self.run_start_time
            if elapsed >= self.seconds:
                return STATE.timeout

        elif self.max_steps:
            if cur_step >= self.max_steps:
                return STATE.max_steps

        elif not self.until_complete:
            if state == STATE.complete:
                # Ignore STATE.complete if until_complete is False
                return STATE.running

        return state



import collections, copy, queue
from . import control
from .. util import artnet_message, log, offset_range, server_cache, udp

QUEUE_TIMEOUT = 0.1


class ArtNet(control.ExtractedControl):
    def __init__(self, *args, ip_address='', port=artnet_message.UDP_PORT,
                 offset=0, **kwds):
        super().__init__(*args, **kwds)
        self.address = ip_address, port
        self.offset = offset_range.DMXChannel.make(offset)

    def _convert(self, msg):
        msg = artnet_message.bytes_to_message(msg)
        if not msg:
            return

        data = bytes(self.offset.read_from(msg.data))
        msg = collections.OrderedDict((
            ('type', 'dmx'),
            ('net', msg.net),
            ('subUni', msg.subUni),
            ('data', data)))

        return super()._convert(msg)

    def _make_thread(self):
        return udp.Receiver(self.address, receive=self.receive)



from . animation import Animation
from .. layout import Matrix


class BaseMatrixAnim(Animation):
    LAYOUT_CLASS = Matrix
    LAYOUT_ARGS = 'width', 'height'

    def __init__(self, layout, width=0, height=0, **kwds):
        super().__init__(layout, **kwds)
        self.width = width or layout.width
        self.height = height or layout.height



import operator, os
from .. util import log

PROJECT_ALIASES = {}

BUILTIN_ALIASES = {
    'apa102': 'bibliopixel.drivers.SPI.APA102.APA102',
    'lpd8806': 'bibliopixel.drivers.SPI.LPD8806.LPD8806',
    'pi_ws281x': 'bibliopixel.drivers.PiWS281X.PiWS281X',
    'serial': 'bibliopixel.drivers.serial.Serial',
    'sk9822': 'bibliopixel.drivers.SPI.APA102.APA102',
    'spi': 'bibliopixel.drivers.SPI.SPI',
    'ws2801': 'bibliopixel.drivers.SPI.WS2801.WS2801',
    'ws281x': 'bibliopixel.drivers.SPI.WS281X.WS281X',

    'reprocess': 'bibliopixel.animation.reprocess.reprocess.Reprocess',
}


def get_alias(alias):
    return PROJECT_ALIASES.get(alias) or BUILTIN_ALIASES.get(alias)



"""
Save the current project defaults to a file
"""

from ... project import defaults


def run(args):
    defaults.save_defaults(args.name)


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.add_argument('name', help='Save the current defaults setting')



from . collection import Collection
from .. util import log


class Indexed(Collection):
    """
    An ``Indexed`` is a :py:class:`bibliopixel.animation.Collection`
    which keeps track of the current animation through an index into the list of
    animations, and gets a callback after that index changes.

    An ``Indexed`` has two properties.

    ``index`` is a mutable property indexing the current animation in the list
    of animations.

    ``current_animation`` returns the animation at position ``index`` or None
    if it is out of the bounds of the ``Collection``.
    """
    _index = -1

    @property
    def index(self):
        """
        :returns int: index of the current animation within the Collection.
        """
        return self._index

    @index.setter
    def index(self, index):
        index = self.animations._index(index)
        self._index, old_index = index, self._index
        self._on_index(old_index)

    def _on_index(self, old_index):
        """
        Override this method to get called right after ``self.index`` is set.

        :param int old_index: the previous index, before it was changed.
        """
        if self.animation:
            log.debug('%s: %s',
                      self.__class__.__name__, self.current_animation.title)
            self.frames = self.animation.generate_frames(False)

    def pre_run(self):
        super().pre_run()
        self.index = 0

    def step(self, amt=1):
        try:
            next(self.frames)
            return True
        except StopIteration:
            pass
        except Exception as e:
            log.error('Exception %s in wrapper.step()', e)

    def forward(self, *unused):
        self.index += 1

    def backward(self, *unused):
        self.index -= 1

    @property
    def animation(self):
        """
        :returns: the selected animation based on self.index, or None if
            self.index is out of bounds
        """
        if 0 <= self._index < len(self.animations):
            return self.animations[self._index]

    @property
    def current_animation(self):
        """
        DEPRECATED:
        :returns: self.animation
        """
        return self.animation
